from collections.abc import Callable
from pathlib import Path
from typing import ClassVar, Final, override

from canvy.const import API_KEY_REGEX, URL_REGEX
from canvy.tui.const import CanvyMode
from canvy.types import CanvyConfig, ModelProvider
from canvy.utils import get_config, set_config
from textual import on
from textual.app import ComposeResult
from textual.containers import HorizontalGroup, VerticalGroup
from textual.screen import Screen
from textual.widgets import Button, Input, OptionList, Static
import logging
import re

from textual.widgets.option_list import Option


logger = logging.getLogger(__name__)

ConfigMutator = Callable[[CanvyConfig], CanvyConfig]


class Provider(Static):
    config: CanvyConfig
    provider_type: ModelProvider

    DEFAULT_CSS: ClassVar[
        str
    ] = """
    Provider {
        layout: vertical;
        width: 100%;
        height: auto;
        border: hkey $primary;
        margin: 1;
    }

    Provider.valid {
        background: rgba(0, 255, 0, 0.16);
        border: hkey green;
    }

    Provider.invalid {
        background: rgba(128, 128, 128, 0.08);
        border: hkey gray;
    }

    Provider > Static {
        width: 100%;
        content-align: center middle;
        text-style: bold;
        margin-bottom: 1;
    }

    Provider Input {
        margin: 0 1;
    }
    """

    @override
    def compose(self) -> ComposeResult:
        yield Static(f"{self.provider_type.value}")
        if self.provider_type is ModelProvider.OPENAI:
            yield Input(
                value=self.config.openai_key,
                placeholder="Enter OpenAI API key",
                id="openai_key",
                password=True,
            )
        elif self.provider_type is ModelProvider.ANTHRO:
            yield Input(
                value=self.config.anthropic_key,
                placeholder="Enter Anthropic API key",
                id="anthropic_key",
                password=True,
            )
        elif self.provider_type is ModelProvider.OLLAMA:
            yield Input(
                value=self.config.ollama_model,
                placeholder="Enter Ollama model name",
                id="ollama_model",
            )

    def __init__(self, config: CanvyConfig, provider: ModelProvider) -> None:
        super().__init__()
        self.config = config
        self.provider_type = provider

    def config_mutator(self, config: CanvyConfig):
        if self.provider_type is ModelProvider.OPENAI:
            config.openai_key = self.config.openai_key
        elif self.provider_type is ModelProvider.ANTHRO:
            config.anthropic_key = self.config.anthropic_key
        elif self.provider_type is ModelProvider.OLLAMA:
            config.ollama_model = self.config.ollama_model
        return config

    def on_mount(self):
        self.refresh_validation_state()

    @property
    def valid(self) -> bool:
        # TODO: Update this
        input_widget = self.query_exactly_one(Input)
        return bool(input_widget.value)

    def refresh_validation_state(self) -> None:
        self.remove_class("valid")
        self.remove_class("invalid")
        self.add_class("valid" if self.valid else "invalid")

    @on(Input.Changed)
    def on_input_changed(self) -> None:
        self.refresh_validation_state()


class Canvas(Static):
    config: Final[CanvyConfig]
    border_title: str

    DEFAULT_CSS: ClassVar[
        str
    ] = """
    Canvas {
        layout: vertical;
        width: 100%;
        height: auto;
        border: hkey $primary;
        margin: 1;
    }

    Canvas.valid {
        background: rgba(0, 255, 0, 0.16);
        border: hkey green;
    }

    Canvas.invalid {
        background: rgba(128, 128, 128, 0.08);
        border: hkey gray;
    }

    Canvas > Static {
        width: 100%;
        content-align: center middle;
        text-style: bold;
        margin-bottom: 1;
    }

    Canvas Input {
        margin: 1;
    }
    """

    def __init__(self, config: CanvyConfig):
        self.config = config
        super().__init__()

    @override
    def compose(self) -> ComposeResult:
        yield Static("Canvas")
        yield Input(self.config.canvas_url, "Canvas URL: ", id="url_input")
        yield Input(
            self.config.canvas_key,
            "Canvas API Key: ",
            id="canvas_key_input",
            password=True,
        )

    def on_mount(self):
        self.refresh_validation_state()

    def config_mutator(self, config: CanvyConfig):
        url_input = self.query_exactly_one("#url_input", expect_type=Input)
        key_input = self.query_exactly_one("#canvas_key_input", expect_type=Input)
        config.canvas_url = url_input.value
        config.canvas_key = key_input.value
        return config

    @property
    def valid(self) -> bool:
        url_input = self.query_exactly_one("#url_input", expect_type=Input)
        url_input = CanvyConfig.add_https(url_input.value)
        key_input = self.query_exactly_one("#canvas_key_input", expect_type=Input).value
        return (
            re.match(CanvyConfig.add_https(URL_REGEX), url_input) is not None
            and re.match(API_KEY_REGEX, key_input) is not None
        )

    @on(Input.Changed)
    def on_input_changed(self) -> None:
        self.refresh_validation_state()

    def refresh_validation_state(self) -> None:
        self.remove_class("valid")
        self.remove_class("invalid")
        self.add_class("valid" if self.valid else "invalid")


class StoragePath(Static):
    config: CanvyConfig
    border_title: str

    DEFAULT_CSS: ClassVar[
        str
    ] = """
    StoragePath {
        layout: vertical;
        width: 100%;
        height: auto;
        border: hkey $primary;
        margin: 1;
    }

    StoragePath.valid {
        background: rgba(0, 255, 0, 0.16);
        border: hkey green;
    }

    StoragePath.invalid {
        background: rgba(128, 128, 128, 0.08);
        border: hkey gray;
    }

    StoragePath > Static {
        width: 100%;
        content-align: center middle;
        text-style: bold;
        margin-bottom: 1;
    }

    StoragePath Input {
        margin: 1;
    }
    """

    @override
    def compose(self) -> ComposeResult:
        yield Static("Storage Path")
        with VerticalGroup() as container:
            yield Input(str(self.config.storage_path), "Path: ", id="path_input")

    def on_mount(self):
        self.refresh_validation_state()

    def __init__(self, config: CanvyConfig):
        self.config = config
        super().__init__()

    @property
    def valid(self):
        path_input = self.query_exactly_one(Input)
        try:
            return CanvyConfig.verify_accessible_path(Path(path_input.value))
        except Exception as e:
            logger.debug(f"Invalid storage provided{e}")
            return False

    def config_mutator(self, config: CanvyConfig):
        path_input = self.query_exactly_one(Input)
        config.storage_path = Path(path_input.value)
        return config

    @on(Input.Changed)
    def on_input_changed(self) -> None:
        self.refresh_validation_state()

    def refresh_validation_state(self) -> None:
        self.remove_class("valid")
        self.remove_class("invalid")
        self.add_class("valid" if self.valid else "invalid")


class DefaultProvider(Static):
    config: CanvyConfig
    selected: ModelProvider
    border_title: str

    DEFAULT_CSS: ClassVar[
        str
    ] = """
    DefaultProvider {
        layout: vertical;
        width: 100%;
        height: auto;
        border: hkey $primary;
        margin: 1;
    }
    """

    @override
    def compose(self) -> ComposeResult:
        yield Static()
        with VerticalGroup():
            yield OptionList(*list(Option(p, id=p) for p in ModelProvider))

    def on_mount(self):
        self.border_title = "Default Provider"

    def __init__(self, config: CanvyConfig):
        self.config = config
        self.selected = config.default_provider
        super().__init__()

    def config_mutator(self, config: CanvyConfig):
        config.default_provider = self.selected
        return config

    @on(OptionList.OptionSelected)
    def update_selected(self, event: OptionList.OptionSelected):
        self.selected = ModelProvider(event.option.id)



class SettingsPage(Screen[None]):
    config: CanvyConfig

    DEFAULT_CSS: ClassVar[
        str
    ] = """
    SettingsPage {
        layout: vertical;
        align: center middle;
    }
    
    #settings_container {
        layout: vertical;
        border-title-align: center;
        max-width: 50%;
        min-width: 60;
        border: heavy $primary;
        padding: 2;
        background: $surface;
    }

    Input {
        margin: 1 0;
    }

    Button {
        margin: 0 2;
    }

    #button_group {
        margin-top: 2;
    }
    """

    def __init__(
        self, name: str | None = None, id: str | None = None, classes: str | None = None
    ) -> None:
        super().__init__(name, id, classes)
        self.config = get_config()

    @override
    def compose(self) -> ComposeResult:
        with VerticalGroup(id="settings_container") as container:
            container.border_title = "Settings"
            yield Canvas(self.config)
            for provider in ModelProvider:
                if provider != ModelProvider.NONE:
                    yield Provider(self.config, provider)
            yield StoragePath(self.config)
            yield DefaultProvider(self.config)
            with HorizontalGroup(id="button_group"):
                yield Button("Save", id="save_button", variant="primary")
                yield Button("Back", id="quit_button", variant="error")

    @on(Button.Pressed, "#save_button")
    def save_settings(self) -> None:
        providers = self.query(Provider)
        canvas = self.query_exactly_one(Canvas)
        storage = self.query_exactly_one(StoragePath)
        default_prov = self.query_exactly_one(DefaultProvider)
        mutations: list[ConfigMutator] = []
        mutations.extend(provider.config_mutator for provider in providers)
        mutations.append(canvas.config_mutator)
        mutations.append(storage.config_mutator)
        mutations.append(default_prov.config_mutator)
        for mu in mutations:
            self.config = mu(self.config)
        try:
            new_config = self.config.model_validate(self.config.model_dump())
            set_config(new_config)
            self.notify("Saved successfully.", severity="information")
        except Exception as e:
            self.notify("Error saving settings!", severity="error")
            logger.info(f"Error saving settings: {e}")
        self.quit()

    @on(Button.Pressed, "#quit_button")
    def quit(self):
        self.app.switch_mode(CanvyMode.MAIN)



from typing import ClassVar, override

from textual.app import ComposeResult
from textual.screen import Screen
from textual.widgets import Button


class TutorPage(Screen[None]):
    DEFAULT_CSS: ClassVar[str] = """"""

    @override
    def compose(self) -> ComposeResult:
        yield Button()



from typing import ClassVar, override

from canvy.tui.const import CanvyMode
from canvy.types import CanvyConfig
from canvy.utils import get_config, provider
from textual import on
from textual.app import ComposeResult
from textual.containers import VerticalGroup
from textual.message import Message
from textual.screen import Screen
from textual.widgets import Button


class MainPage(Screen[None]):
    DEFAULT_CSS: ClassVar[
        str
    ] = """
    MainPage {
        align: center middle;
    }

    VerticalGroup {
        align: center top;
        border: heavy gray;
        max-width: 15%;
    }

    Button {
        margin: 1 0;
    }

    .disabled {
        hatch: cross gray;
    }
    """

    config: CanvyConfig

    class Switch(Message):
        def __init__(self, target: CanvyMode) -> None:
            self.target: str = target
            super().__init__()

    def __init__(
        self, name: str | None = None, id: str | None = None, classes: str | None = None
    ) -> None:
        self.config = get_config()
        super().__init__(name, id, classes)

    @override
    def compose(self) -> ComposeResult:
        # TODO: Check if this raises an error on shorthand syntax
        with VerticalGroup():
            yield Button("Download", id="download_button")
            yield Button(
                "Tutor", id="tutor_button", disabled=provider(self.config) is None
            )
            yield Button("Settings", id="settings_button")
            yield Button("Quit", id="quit_button", variant="error")

    @on(Button.Pressed)
    def emit_switch(self, event: Button.Pressed):
        selected = {
            "download_button": CanvyMode.DOWNLOAD,
            "tutor_button": CanvyMode.TUTOR,
            "settings_button": CanvyMode.SETTINGS
        }.get(event.button.id or "")
        self.app.switch_mode(selected or CanvyMode.MAIN)

    @on(Button.Pressed, "#quit_button")
    def quit(self):
        self.app.exit()



from __future__ import annotations

import logging.config
import os
import re
from collections.abc import Iterable
from functools import reduce
from pathlib import Path

from agno.models.ollama import Ollama
from agno.models.openai.chat import OpenAIChat
from agno.models.anthropic import Claude
import toml
from canvasapi.file import File

from canvy.const import (
    ANTHRO_MODEL,
    CONFIG_PATH,
    LOG_FN,
    LOGGING_CONFIG,
    OPENAI_MODEL,
    SUMMARIES_DIRNAME,
)
from canvy.types import CanvyConfig, Model

logger = logging.getLogger(__name__)


def setup_logging() -> None:
    """
    Setup logging using logging config defined in const.py because it's
    quite good
    """
    create_dir(LOG_FN.parent)
    return logging.config.dictConfig(LOGGING_CONFIG)


def better_course_name(name: str) -> str:
    """
    Removes ID numbers next to the given title of the course
    """
    return re.sub(r" ?\((\d,? ?)+\) ?", "", name)


def create_dir(directory: Path) -> None:
    logger.debug(f"Creating directory {directory} if not existing")
    os.makedirs(directory, exist_ok=True)


def has_config(path: Path | None = None) -> bool:
    try:
        get_config(path)
        return True
    except Exception as e:
        logger.info(f"Config test fail: {e}")
        return False


def get_config(path: Path | None = None) -> CanvyConfig:
    path = path or CONFIG_PATH
    with open(path) as fp:
        logger.debug(f"Retrieving config from {path}")
        config = CanvyConfig(**toml.load(fp))  # pyright: ignore[reportAny]
    return config


def set_config(config: CanvyConfig, dest: Path = CONFIG_PATH) -> None:
    dest = dest if dest else CONFIG_PATH
    with open(dest, "w") as fp:
        logger.debug("Writing config")
        toml.dump(config.model_dump(), fp)


def delete_config(path: Path = CONFIG_PATH):
    os.remove(path)


def download_structured(
    file: File, *dirs: str, storage_dir: Path | None = None, force: bool = False
) -> bool:
    """
    Download a canvasapi File and preserve course structure using directory names

    Args:
        file: File object given by Canvas, can raise various exceptions
        dirs: Series of directory names to make and download file into
        force: Overwrite any previously existing files

    Returns:
        If the file was downloaded
    """
    download_dir = Path(storage_dir or get_config().storage_path).expanduser()
    file_name = file.filename  # pyright: ignore[reportAny]
    file_path: Path = concat_names(download_dir, (*dirs, file_name))
    create_dir(file_path.parent)
    if not file_path.is_file() or force:
        logger.info(f"Downloading {file_name}{'(forced)' * force} into {file_path}")
        try:
            file.download(file_path)  # pyright: ignore[reportUnknownMemberType]
            return True
        except Exception as e:
            logger.warning(
                f"Tried to download {file_name} but we likely don't have access ({e})"
            )
            return False
    else:
        logger.info(f"{file_name} already present, skipping")
        return False


def concat_names(base: Path, names: Iterable[str | Path]) -> Path:
    return reduce(lambda p, q: p / q, [base, *map(Path, names)])


def provider(config: CanvyConfig) -> Model:
    """
    Get the preferred model provider from the config, default is OpenAI because
    lazy people. Implemented config check to prevent ambiguous errors

    Args:
        config: Config with provider

    Returns:
        Model: Agno model type to use, of which there are many
    """
    if config.default_provider == "OpenAI":
        if key := config.openai_key:
            return OpenAIChat(id=OPENAI_MODEL, api_key=key)
    elif config.default_provider == "Anthropic":
        if key := config.anthropic_key:
            return Claude(id=ANTHRO_MODEL, api_key=key)
    elif config.default_provider == "Ollama":
        if model := config.ollama_model:
            return Ollama(id=model)
    return None


def setup_cache_mirror(config: CanvyConfig):
    base = config.storage_path / SUMMARIES_DIRNAME
    logger.info(f"Setting up cache mirror at {base}")
    for path in config.storage_path.rglob("*"):
        rebased = base / (path.relative_to(config.storage_path))
        logger.info(f"{path=} {rebased=}")
        if SUMMARIES_DIRNAME in str(path):
            continue
        elif path.is_file():
            rebased.parent.mkdir(parents=True, exist_ok=True)
            rebased.touch()
        elif path.is_dir():
            rebased.mkdir(parents=True, exist_ok=True)


def put_summary(config: CanvyConfig, path: Path, summary: str):
    path = (
        config.storage_path / SUMMARIES_DIRNAME / path.relative_to(config.storage_path)
    )
    path.write_text(summary)
    logger.info(f"Put summary {summary[:17]}.. into {path}")


def get_summary(config: CanvyConfig, path: Path):
    path = (
        config.storage_path / SUMMARIES_DIRNAME / path.relative_to(config.storage_path)
    )
    content = path.read_text()
    logger.info(f"Read summary {content}.. from {path}")
    return content



# INFO: Postpone but not delete code because idk why I can't access these marks
import logging

from canvasapi.assignment import Assignment
from canvasapi.canvas import Canvas

logger = logging.getLogger(__name__)


def grades_by_course(_canvas: Canvas) -> dict[str, float | None]:
    ...
    # return {
    #     course.course_code: calculate_grading(assignment)
    #     for course in canvas.get_courses()
    #     for assignment in course.get_assignments()
    # }


def grades(_canvas: Canvas) -> dict[str, dict[str, float | None]]:
    ...
    # return {
    #     course.course_code: {
    #         assignment.name: calculate_grading(assignment)
    #         for assignment in course.get_assignments()
    #     }
    #     for course in canvas.get_courses()
    # }


def calculate_grading(_assignment: Assignment) -> float | None:
    ...
    # def grade_change_value(event: GradeChangeEvent) -> float:
    #     t = datetime.fromisoformat(event.created_at)
    #     return t.timestamp()

    # logger.info("Getting grade change events")
    # try:
    #     grade_events = list(assignment.get_grade_change_events())
    # except Exception as e:
    #     logger.warning(f"Error fetching grade events: {e}")
    #     return None
    # if not grade_events:
    #     logger.warning(f"Grade events for {assignment.name} are empty!")
    #     return None
    # grade_events.sort(key=grade_change_value)
    # last_event = grade_events.pop()
    # return last_event.grade_after / assignment.points_possible



from typing import ClassVar, override

from canvy.types import CanvyConfig
from canvy.utils import get_config
from textual import on
from textual.app import ComposeResult
from textual.containers import HorizontalGroup
from textual.screen import Screen
from textual.widgets import Button, Input


class HelpPage(Screen[None]):
    DEFAULT_CSS: ClassVar[str] = """
    Help {
        layout: vertical;
        align: center middle;
        border-title-align: center;
        max-width: 30%;
        border: heavy gray;
        padding: 2
    }

    Input {
        margin: 1;
    }

    Button  {
        margin: 1;
    }
    """

    config: CanvyConfig

    def __init__(self, name: str | None = None, id: str | None = None, classes: str | None = None) -> None:
        super().__init__(name, id, classes)
        self.config = get_config()

    @override
    def compose(self) -> ComposeResult:
        yield Input("", id="")
        with HorizontalGroup():
            yield Button("Save", id="save_button", variant="primary")
            yield Button("Back", id="quit_button", variant="error")

    @on(Button.Pressed, "#quit_button")
    def quit(self):
        self.app.exit()



import logging
from typing import ClassVar, override

from canvy.tui.const import CanvyMode
from pydantic import ValidationError
from textual import on
from textual.app import ComposeResult
from textual.containers import Center
from textual.screen import Screen
from textual.widgets import Button, Input

from canvy.const import LOGIN_CULPRITS
from canvy.types import CanvyConfig
from canvy.utils import set_config

logger = logging.getLogger(__name__)


class LoginPage(Screen[None]):
    DEFAULT_CSS: ClassVar[
        str
    ] = """
    Screen {
        align: center middle;
    }

    LoginPage {
        layout: vertical;
        align: center middle;
        border-title-align: center;
        transition: offset 200ms;
        width: 30%;
        border: heavy gray;
        padding: 2;
    }

    Input {
        margin: 1;
    }

    Button {
        margin: 1;
    }
    """

    @override
    def compose(self) -> ComposeResult:
        yield Input(placeholder="Your Canvas URL", id="url_input")
        yield Input(placeholder="API Key", id="sk_input")
        with Center():
            yield Button("Submit", variant="success")

    def on_mount(self):
        self.border_title: str = "Login"

    @on(Button.Pressed)
    def validate_inputs(self, _: Button.Pressed):
        url_input: str = self.query_one("#url_input", expect_type=Input).value
        sk_input: str = self.query_one("#sk_input", expect_type=Input).value
        logger.info(f"Login page info: {url_input}: {sk_input[:5] + "."*20}")
        try:
            config = CanvyConfig(canvas_url=url_input, canvas_key=sk_input)
            set_config(config)
            self.app.switch_mode(CanvyMode.MAIN)
        except ValidationError as e:
            logger.info(f"Invalid login page submission: {e}")
            culprits: list[str] = [str(next(iter(d["loc"]))) for d in e.errors()]
            for culprit in reversed(culprits):
                self.notify(f"Invalid [b]{LOGIN_CULPRITS[culprit]}[/b]", severity="error")



# pyright: reportAny=false
# pyright: reportUnknownVariableType=false
# pyright: reportUnknownArgumentType=false
# pyright: reportUnknownMemberType=false
import logging
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from threading import Lock
from typing import ClassVar, override

from agno.document.reader.pdf_reader import PDFReader
from canvasapi.canvas import Canvas
from canvasapi.file import File
from canvy.tui.const import CanvyMode
from textual import on, work
from textual.app import ComposeResult
from textual.color import Gradient
from textual.containers import HorizontalGroup, VerticalGroup, VerticalScroll
from textual.message import Message
from textual.reactive import reactive
from textual.screen import Screen
from textual.timer import Timer
from textual.widgets import (
    Button,
    DirectoryTree,
    Label,
    Markdown,
    ProgressBar,
)

from canvy.const import DOCSCRAPE_DEFAULT_MSG
from canvy.scripts import tutor
from canvy.scripts.downloader import module_item_files
from canvy.types import CanvyConfig, Model
from canvy.utils import (
    download_structured,
    get_config,
    get_summary,
    provider,
    put_summary,
    setup_cache_mirror,
)

logger = logging.getLogger(__name__)


COOL_GRADIENT: Gradient = Gradient.from_colors(
    "#881177",
    "#aa3355",
    "#cc6666",
    "#ee9944",
    "#eedd00",
    "#99dd55",
    "#44dd88",
    "#22ccbb",
    "#00bbcc",
    "#0099cc",
    "#3366bb",
    "#663399",
)


class FSTree(DirectoryTree):
    timer: Timer | None

    DEFAULT_CSS: ClassVar[
        str
    ] = """
    FSTree {
        width: 35%;
        border: vkey gray;
        background: $background;
    }
    """

    def __init__(
        self,
        path: str | Path,
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        self.timer = None
        super().__init__(path, name=name, id=id, classes=classes, disabled=disabled)

    @override
    def on_mount(self):
        super().on_mount()
        self.timer = self.set_interval(3, callback=self.update_time)

    def update_time(self):
        self.reload()


class Content(VerticalGroup):
    """
    Show extracted PDF contents for fun
    """

    DEFAULT_CSS: ClassVar[
        str
    ] = """
    Content {
        content-align: center top;
        height: 100%
    }
    /*
    border: 'ascii', 'blank', 'dashed', 'double', 'heavy', 'hidden', 'hkey', 'inner',
            'none', 'outer', 'panel', 'round', 'solid', 'tab', 'tall', 'thick', 'vkey',
            'wide'
    */
    """

    @override
    def compose(self) -> ComposeResult:
        yield VerticalScroll(
            Markdown(
                """\
# Canvy

Content for a selected file will appear here, it doesn't work very well with PDFs at the \
moment - unless you use an LLM.

## Supported File Types

Markdown syntax and extensions are supported.

- PDF documents
- Plaintext
- Markdown
"""
            )
        )


class DownloadControl(HorizontalGroup):
    """
    Report download progress
    """

    DEFAULT_CSS: ClassVar[
        str
    ] = """
    DownloadControl {
        dock: bottom;
        align: center middle;
        border: hkey gray;
    }

    #group_1 {
        align: center middle;
        width: 65%;
    }

    #group_2 {
        align: center middle;
        width: 35%;
    }

    #download_button {
        margin-right: 3;
    }

    #cancel_button {
        margin-right: 3;
    }

    VerticalGroup {
        width: auto;
        margin: 0 3;
    }
    """

    download_count: reactive[int] = reactive(0)
    _terminate: bool = False

    @property
    def terminate(self):
        return self._terminate

    @override
    def compose(self) -> ComposeResult:
        with HorizontalGroup(id="group_1"):
            with VerticalGroup():
                yield Label("Files", id="lab_files")
                yield ProgressBar(id="pg_files", show_eta=False, gradient=COOL_GRADIENT)
            with VerticalGroup():
                yield Label("Modules", id="lab_modules")
                yield ProgressBar(
                    id="pg_modules", show_eta=False, gradient=COOL_GRADIENT
                )
            with VerticalGroup():
                yield Label("Courses", id="lab_courses")
                yield ProgressBar(
                    id="pg_courses", show_eta=False, gradient=COOL_GRADIENT
                )
        with HorizontalGroup(id="group_2"):
            yield Button("Download", id="download_button", variant="success")
            yield Button("Cancel", id="cancel_button", variant="primary")
            yield Button("Back", id="quit_button", variant="error")

    class Start(Message): ...

    class Stop(Message): ...

    @on(Button.Pressed, "#download_button")
    def start_download(self) -> None:
        self.post_message(self.Start())
        self.download()

    @on(Button.Pressed, "#cancel_button")
    def stop_download(self) -> None:
        self._terminate = True
        self.query_exactly_one("#pg_courses", expect_type=ProgressBar).update(total=0)
        self.query_exactly_one("#pg_modules", expect_type=ProgressBar).update(total=0)
        self.query_exactly_one("#pg_files", expect_type=ProgressBar).update(total=0)
        self.post_message(self.Stop())

    @on(Button.Pressed, "#quit_button")
    def quit(self) -> None:
        self.app.switch_mode(CanvyMode.MAIN)

    @work(thread=True)
    def download(self, *, force: bool = False):
        # TODO: Fix course length inaccuracy
        count_lock = Lock()
        config = get_config()
        canvas = Canvas(config.canvas_url, config.canvas_key)

        def safe_download(file: File, paths: list[str]):
            def inner():
                res = download_structured(
                    file, *paths, storage_dir=config.storage_path, force=force
                )
                with count_lock:
                    self.download_count += res

            return inner

        with ThreadPoolExecutor(max_workers=5) as executor:
            progress_courses = self.query_exactly_one(
                "#pg_courses", expect_type=ProgressBar
            )
            progress_modules = self.query_exactly_one(
                "#pg_modules", expect_type=ProgressBar
            )
            progress_files = self.query_exactly_one(
                "#pg_files", expect_type=ProgressBar
            )
            label_courses = self.query_exactly_one("#lab_courses", expect_type=Label)
            label_modules = self.query_exactly_one("#lab_modules", expect_type=Label)
            label_files = self.query_exactly_one("#lab_files", expect_type=Label)
            user_courses = list(canvas.get_courses(enrollment_state="active"))
            for course in user_courses:
                label_courses.update(f"Course: {course.course_code}")
                progress_courses.update(total=len(user_courses))
                for module in (modules := list(course.get_modules())):
                    label_modules.update(f"Module: {module.name}")
                    progress_modules.update(total=len(modules))
                    for item in module.get_module_items():
                        path_files = list(
                            module_item_files(canvas, course, module, item)
                        )
                        progress_files.update(total=len(path_files))
                        for paths, file in path_files:
                            if self.terminate:
                                executor.shutdown(wait=False, cancel_futures=True)
                                return
                            else:
                                label_files.update(f"Files: {file.filename}")
                                executor.submit(safe_download(file, paths))
                        progress_files.advance()
                    progress_modules.advance()
                    progress_files.update(total=0)
                progress_courses.advance()
                progress_modules.update(total=0)
            progress_courses.update(total=0)


class DownloadPage(Screen[None]):
    llm_provider: Model
    config: CanvyConfig

    def __init__(
        self, name: str | None = None, id: str | None = None, classes: str | None = None
    ) -> None:
        super().__init__(name, id, classes)
        self.config = get_config()
        self.llm_provider = provider(self.config)
        if self.llm_provider is not None:
            setup_cache_mirror(self.config)  # maybe expensive?

    @override
    def compose(self) -> ComposeResult:
        with HorizontalGroup():
            fst = FSTree(self.config.storage_path)
            yield fst
            yield Content()
        yield DownloadControl()

    @on(FSTree.FileSelected)
    def update_on_select(self, msg: FSTree.FileSelected):
        logger.info(f"File {msg.path} selected on FSTree")
        file_path = msg.path
        md_widget = self.query_exactly_one(Markdown)
        if (ext := file_path.suffix) == ".pdf":
            self.populate_document(file_path)
        elif ext in {".txt", ".md", ""}:
            md_widget.update(file_path.read_text())
        else:
            logger.warning(f"Unhandled file type: {file_path}")

    def populate_document(self, file_path: Path):
        md_widget = self.query_exactly_one(Markdown)
        documents = PDFReader().read(file_path)
        content = "".join(d.content for d in documents)
        if self.llm_provider is None:
            md_widget.update(DOCSCRAPE_DEFAULT_MSG + content)
        else:
            agent = tutor(self.config, interactive=False)
            if response := get_summary(self.config, file_path):
                md_widget.update(response)
                return
            response = agent.run(
                f"Summarise the content from: {file_path}", stream=True
            )
            total = []
            for r in response:
                # TODO: Re-rendering causes the FSTree to not populate while the response is being built
                total.append(r.content)
                md_widget.update("".join(total))
            put_summary(self.config, file_path, "".join(total))

    def on_mount(self):
        self.border_title: str = "Login"



import logging
from collections.abc import Callable
from typing import ClassVar

from canvy.tui.const import CanvyMode
from canvy.tui.help import HelpPage
from canvy.tui.main import MainPage
from canvy.tui.settings import SettingsPage
from canvy.tui.tutor import TutorPage
from textual.app import App
from textual.binding import BindingType
from textual.screen import Screen

from canvy.tui.download import DownloadPage
from canvy.tui.login import LoginPage
from canvy.utils import has_config

logger = logging.getLogger(__name__)


class Canvy(App[None]):
    MODES: ClassVar[dict[str, str | Callable[..., Screen[None]]]] = {
        CanvyMode.LOGIN: LoginPage,
        CanvyMode.DOWNLOAD: DownloadPage,
        CanvyMode.TUTOR: TutorPage,
        CanvyMode.SETTINGS: SettingsPage,
        CanvyMode.MAIN: MainPage,
        CanvyMode.HELP: HelpPage,
    }
    BINDINGS: ClassVar[list[BindingType]] = []

    def on_mount(self):
        self.switch_mode("login" if not has_config() else "main")


def run():
    app = Canvy()
    app.run()
